#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from pathlib import Path
from typing import Iterable, Optional, Union

from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.spider_max import SPIDErMaxOuts, spider_max

pylog = logging.getLogger(__name__)


class SPIDErMax(AACMetric[Union[SPIDErMaxOuts, Tensor]]):
    """SPIDEr-max class.

    - Paper: https://hal.archives-ouvertes.fr/hal-03810396/file/Labbe_DCASE2022.pdf

    For more information, see :func:`~aac_metrics.functional.spider.spider`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = 0.0
    max_value = 5.5

    def __init__(
        self,
        return_all_scores: bool = True,
        *,
        return_all_cands_scores: bool = False,
        # CIDEr args
        n: int = 4,
        sigma: float = 6.0,
        # SPICE args
        cache_path: Union[str, Path, None] = None,
        java_path: Union[str, Path, None] = None,
        tmp_path: Union[str, Path, None] = None,
        n_threads: Optional[int] = None,
        java_max_memory: str = "8G",
        timeout: Union[None, int, Iterable[int]] = None,
        verbose: int = 0,
    ) -> None:
        super().__init__()
        self._return_all_scores = return_all_scores
        self._return_all_cands_scores = return_all_cands_scores
        self._n = n
        self._sigma = sigma
        self._cache_path = cache_path
        self._java_path = java_path
        self._tmp_path = tmp_path
        self._n_threads = n_threads
        self._java_max_memory = java_max_memory
        self._timeout = timeout
        self._verbose = verbose

        self._mult_candidates = []
        self._mult_references = []

    def compute(self) -> Union[SPIDErMaxOuts, Tensor]:
        return spider_max(
            mult_candidates=self._mult_candidates,
            mult_references=self._mult_references,
            return_all_scores=self._return_all_scores,
            return_all_cands_scores=self._return_all_cands_scores,
            n=self._n,
            sigma=self._sigma,
            cache_path=self._cache_path,
            java_path=self._java_path,
            tmp_path=self._tmp_path,
            n_threads=self._n_threads,
            java_max_memory=self._java_max_memory,
            timeout=self._timeout,
            verbose=self._verbose,
        )

    def extra_repr(self) -> str:
        hparams = {
            "n": self._n,
            "sigma": self._sigma,
            "java_max_memory": self._java_max_memory,
        }
        repr_ = ", ".join(f"{k}={v}" for k, v in hparams.items())
        return repr_

    def get_output_names(self) -> tuple[str, ...]:
        output_names = ["spider_max"]
        if self._return_all_cands_scores:
            output_names += ["cider_d_all", "spice_all", "spider_all"]
        return tuple(output_names)

    def reset(self) -> None:
        self._mult_candidates = []
        self._mult_references = []
        return super().reset()

    def update(
        self,
        mult_candidates: list[list[str]],
        mult_references: list[list[str]],
    ) -> None:
        self._mult_candidates += mult_candidates
        self._mult_references += mult_references



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Optional, Union, get_args

import torch
from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.clap_sim import (
    CLAP,
    DEFAULT_CLAP_SIM_MODEL,
    CLAPMethod,
    CLAPOuts,
    _load_clap,
    clap_sim,
)
from aac_metrics.utils.globals import _get_device


class CLAPSim(AACMetric[Union[CLAPOuts, Tensor]]):
    """Cosine-similarity of the Contrastive Language-Audio Pretraining (CLAP) embeddings.

    The implementation is based on the msclap pypi package.
    Note: Instances of this class are not pickable.

    - Paper: https://arxiv.org/pdf/2411.00321
    - msclap package: https://pypi.org/project/msclap/

    For more information, see :func:`~aac_metrics.functional.clap_sim.clap_sim`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = -1.0
    max_value = 1.0

    def __init__(
        self,
        return_all_scores: bool = True,
        *,
        clap_method: CLAPMethod = "text",
        clap_model: Union[str, CLAP] = DEFAULT_CLAP_SIM_MODEL,
        device: Union[str, torch.device, None] = "cuda_if_available",
        batch_size: Optional[int] = 32,
        reset_state: bool = True,
        seed: Optional[int] = 42,
        verbose: int = 0,
    ) -> None:
        if clap_method not in get_args(CLAPMethod):
            msg = f"Invalid argument {clap_method=}. (expected one of {get_args(CLAPMethod)})"
            raise ValueError(msg)

        device = _get_device(device)
        clap_model = _load_clap(
            clap_model=clap_model,
            device=device,
            reset_state=reset_state,
        )

        super().__init__()
        self._return_all_scores = return_all_scores
        self._clap_method: CLAPMethod = clap_method
        self._clap_model = clap_model
        self._device = device
        self._batch_size = batch_size
        self._reset_state = reset_state
        self._seed = seed
        self._verbose = verbose

        self._candidates = []
        self._mult_references = []
        self._audio_paths = []

    def compute(self) -> Union[CLAPOuts, Tensor]:
        return clap_sim(
            candidates=self._candidates,
            mult_references=self._mult_references,
            audio_paths=self._audio_paths,
            clap_method=self._clap_method,
            return_all_scores=self._return_all_scores,
            clap_model=self._clap_model,
            device=self._device,
            batch_size=self._batch_size,
            reset_state=self._reset_state,
            seed=self._seed,
            verbose=self._verbose,
        )

    def extra_repr(self) -> str:
        if isinstance(self._clap_model, str):
            model_name = self._clap_model
        else:
            model_name = self._clap_model.__class__.__name__

        hparams = {"clap_model": model_name}
        repr_ = ", ".join(f"{k}={v}" for k, v in hparams.items())
        return repr_

    def get_output_names(self) -> tuple[str, ...]:
        return ("clap_sim",)

    def reset(self) -> None:
        self._candidates = []
        self._mult_references = []
        self._audio_paths = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references_or_audio_paths: Union[list[list[str]], list[str]],
    ) -> None:
        self._candidates += candidates
        if self._clap_method == "audio":
            self._mult_references += mult_references_or_audio_paths
        elif self._clap_method == "text":
            self._audio_paths += mult_references_or_audio_paths
        else:
            msg = f"Invalid value {self._clap_method=}. (expected one of {get_args(CLAPMethod)})"
            raise ValueError(msg)

    def __getstate__(self) -> Any:
        raise RuntimeError(f"{self.__class__.__name__} is not pickable.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pickle
import unittest
from pickle import PicklingError
from unittest import TestCase

from aac_metrics.classes import MACE, CLAPSim, Evaluate


class TestPickle(TestCase):
    # Tests methods
    def test_pickle_dump(self) -> None:
        metrics = Evaluate(metrics="all")

        # Exclude non-pickable metrics
        for i, metric in enumerate(metrics):
            if isinstance(metric, (CLAPSim, MACE)):
                metrics.pop(i)

        try:
            pickle.dumps(metrics)
        except PicklingError as err:
            msg = f"Cannot pickle {metrics.__class__.__name__}.\n{err}"
            self.assertTrue(False, msg)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from typing import Any, Literal, MutableMapping, Optional, TypedDict, Union, get_args

import torch
from msclap import CLAP
from torch import Tensor
from transformers.models.auto.tokenization_auto import AutoTokenizer

from aac_metrics.functional.clap_sim import (
    DEFAULT_CLAP_SIM_MODEL,
    CLAPOuts,
    _load_clap,
    clap_sim,
)
from aac_metrics.functional.fer import (
    DEFAULT_FER_MODEL,
    BERTFlatClassifier,
    FEROuts,
    _load_echecker_and_tokenizer,
    fer,
)

pylog = logging.getLogger(__name__)

MACEMethod = Literal["text", "audio", "combined"]

MACEScores = TypedDict(
    "MACEScores", {"mace": Tensor, "fer": Tensor, "clap_sim": Tensor}
)
MACEOuts = tuple[MACEScores, MACEScores]


def mace(
    candidates: list[str],
    mult_references: Optional[list[list[str]]] = None,
    audio_paths: Optional[list[str]] = None,
    return_all_scores: bool = True,
    *,
    mace_method: MACEMethod = "text",
    penalty: float = 0.3,
    # CLAP args
    clap_model: Union[str, CLAP] = DEFAULT_CLAP_SIM_MODEL,
    seed: Optional[int] = 42,
    # FER args
    echecker: Union[str, BERTFlatClassifier] = DEFAULT_FER_MODEL,
    echecker_tokenizer: Optional[AutoTokenizer] = None,
    error_threshold: float = 0.97,
    device: Union[str, torch.device, None] = "cuda_if_available",
    batch_size: Optional[int] = 32,
    reset_state: bool = True,
    return_probs: bool = False,
    # Other args
    verbose: int = 0,
) -> Union[Tensor, MACEOuts]:
    """Multimodal Audio-Caption Evaluation class (MACE).

    MACE is a metric designed for evaluating automated audio captioning (AAC) systems.
    Unlike metrics that compare machine-generated captions solely to human references, MACE uses both audio and text to improve evaluation.
    By integrating both audio and text, it produces assessments that align better with human judgments.

    The implementation is based on the mace original implementation (original author have accepted to include their code in aac-metrics under the MIT license).

    - Paper: https://arxiv.org/pdf/2411.00321
    - Original author: Satvik Dixit
    - Original implementation: https://github.com/satvik-dixit/mace/tree/main

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target when method is "text" or "combined". defaults to None.
    :param audio_paths: Audio filepaths required when method is "audio" or "combined". defaults to None.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param mace_method: The method used to encode the sentences. Can be "text", "audio" or "combined". defaults to "text".
    :param penalty: The penalty coefficient applied. Higher value means to lower the cos-sim scores when an error is detected. defaults to 0.3.
    :param clap_model: The CLAP model used to extract CLAP embeddings for cosine-similarity. defaults to "MS-CLAP-2023".
    :param seed: Optional seed to make CLAP-sim scores deterministic when using mace_method="audio" or "combined" on large audio files. defaults to 42.
    :param echecker: The echecker model used to detect fluency errors.
        Can be "echecker_clotho_audiocaps_base", "echecker_clotho_audiocaps_tiny", "none" or None.
        defaults to "echecker_clotho_audiocaps_base".
    :param echecker_tokenizer: The tokenizer of the echecker model.
        If None and echecker is not None, this value will be inferred with `echecker.model_type`.
        defaults to None.
    :param error_threshold: The threshold used to detect fluency errors for echecker model. defaults to 0.97.
    :param device: The PyTorch device used to run pre-trained models. If "cuda_if_available", it will use cuda if available. defaults to "cuda_if_available".
    :param batch_size: The batch size of the CLAP and echecker models. defaults to 32.
    :param reset_state: If True, reset the state of the PyTorch global generator after the initialization of the pre-trained models. defaults to True.
    :param return_probs: If True, return each individual error probability given by the fluency detector model. defaults to False.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    # Init models
    clap_model, echecker, echecker_tokenizer = _load_models_and_tokenizer(
        clap_model=clap_model,
        echecker=echecker,
        echecker_tokenizer=echecker_tokenizer,
        device=device,
        reset_state=reset_state,
        verbose=verbose,
    )

    clap_kwds: dict[str, Any] = dict(
        candidates=candidates,
        mult_references=mult_references,
        audio_paths=audio_paths,
        return_all_scores=True,
        clap_model=clap_model,
        device=device,
        batch_size=batch_size,
        reset_state=reset_state,
        seed=seed,
        verbose=verbose,
    )
    if mace_method == "text":
        clap_sim_outs: CLAPOuts = clap_sim(clap_method=mace_method, **clap_kwds)  # type: ignore

    elif mace_method == "audio":
        clap_sim_outs: CLAPOuts = clap_sim(clap_method=mace_method, **clap_kwds)  # type: ignore

    elif mace_method == "combined":
        clap_sim_outs_text: CLAPOuts = clap_sim(  # type: ignore
            clap_method="text", **clap_kwds
        )
        clap_sim_outs_audio: CLAPOuts = clap_sim(  # type: ignore
            clap_method="audio", **clap_kwds
        )

        clap_sim_outs_corpus_text, clap_sim_outs_sents_text = clap_sim_outs_text
        clap_sim_outs_corpus_audio, clap_sim_outs_sents_audio = clap_sim_outs_audio
        clap_sim_outs_corpus = _average_dicts(
            clap_sim_outs_corpus_text, clap_sim_outs_corpus_audio  # type: ignore
        )
        clap_sim_outs_sents = _average_dicts(
            clap_sim_outs_sents_text, clap_sim_outs_sents_audio  # type: ignore
        )
        clap_sim_outs: CLAPOuts = clap_sim_outs_corpus, clap_sim_outs_sents  # type: ignore

    else:
        msg = (
            f"Invalid argument {mace_method=}. (expected one of {get_args(MACEMethod)})"
        )
        raise ValueError(msg)

    fer_outs: FEROuts = fer(  # type: ignore
        candidates=candidates,
        return_all_scores=True,
        echecker=echecker,
        echecker_tokenizer=echecker_tokenizer,
        error_threshold=error_threshold,
        device=device,
        batch_size=batch_size,
        reset_state=reset_state,
        return_probs=return_probs,
        verbose=verbose,
    )
    mace_outs = _mace_from_outputs(clap_sim_outs, fer_outs, penalty)

    if return_all_scores:
        return mace_outs
    else:
        return mace_outs[0]["mace"]


def _average_dicts(
    dict1: MutableMapping[str, Tensor],
    dict2: MutableMapping[str, Tensor],
) -> dict[str, Tensor]:
    averaged_dict = {}
    for key in dict1:
        averaged_dict[key] = (dict1[key] + dict2[key]) / 2
    return averaged_dict  # type: ignore


def _mace_from_outputs(
    clap_sim_outs: CLAPOuts,
    fer_outs: FEROuts,
    penalty: float = 0.3,
) -> MACEOuts:
    """Combines CLAP and FER outputs.

    Based on https://github.com/blmoistawinde/fense/blob/main/fense/evaluator.py#L121
    """
    clap_sim_outs_corpus, clap_sim_outs_sents = clap_sim_outs
    fer_outs_corpus, fer_outs_sents = fer_outs

    clap_sims_scores = clap_sim_outs_sents["clap_sim"]
    fer_scores = fer_outs_sents["fer"]
    mace_scores = clap_sims_scores * (1.0 - penalty * fer_scores)
    # note: we use numpy mean to keep the same values than the original mace, this is only for backward compatibility
    mace_score = torch.as_tensor(
        mace_scores.cpu().numpy().mean(),
        device=mace_scores.device,
    )

    mace_outs_corpus = clap_sim_outs_corpus | fer_outs_corpus | {"mace": mace_score}  # type: ignore
    mace_outs_sents = clap_sim_outs_sents | fer_outs_sents | {"mace": mace_score}  # type: ignore
    mace_outs = mace_outs_corpus, mace_outs_sents

    return mace_outs


def _load_models_and_tokenizer(
    clap_model: Union[str, CLAP] = DEFAULT_CLAP_SIM_MODEL,
    echecker: Union[str, BERTFlatClassifier] = DEFAULT_FER_MODEL,
    echecker_tokenizer: Optional[AutoTokenizer] = None,
    device: Union[str, torch.device, None] = "cuda_if_available",
    reset_state: bool = True,
    verbose: int = 0,
) -> tuple[CLAP, BERTFlatClassifier, AutoTokenizer]:
    clap_model = _load_clap(
        clap_model=clap_model,
        device=device,
        reset_state=reset_state,
    )
    echecker, echecker_tokenizer = _load_echecker_and_tokenizer(
        echecker=echecker,
        echecker_tokenizer=echecker_tokenizer,
        device=device,
        reset_state=reset_state,
        verbose=verbose,
    )
    return clap_model, echecker, echecker_tokenizer



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Metrics for evaluating Automated Audio Captioning systems, designed for PyTorch. """


__author__ = "Étienne Labbé (Labbeti)"
__author_email__ = "labbeti.pub@gmail.com"
__docs__ = "Audio Captioning Metrics"
__docs_url__ = "https://aac-metrics.readthedocs.io/en/stable/"
__license__ = "MIT"
__maintainer__ = "Étienne Labbé (Labbeti)"
__name__ = "aac-metrics"
__status__ = "Development"
__version__ = "0.6.0"


from .classes.base import AACMetric
from .classes.bert_score_mrefs import BERTScoreMRefs
from .classes.bleu import BLEU, BLEU1, BLEU2, BLEU3, BLEU4
from .classes.cider_d import CIDErD
from .classes.clap_sim import CLAPSim
from .classes.evaluate import (
    DCASE2023Evaluate,
    DCASE2024Evaluate,
    Evaluate,
    _get_metric_factory_classes,
    _instantiate_metrics_classes,
)
from .classes.fense import FENSE
from .classes.fer import FER
from .classes.mace import MACE
from .classes.meteor import METEOR
from .classes.rouge_l import ROUGEL
from .classes.sbert_sim import SBERTSim
from .classes.spice import SPICE
from .classes.spider import SPIDEr
from .classes.spider_fl import SPIDErFL
from .classes.spider_max import SPIDErMax
from .classes.vocab import Vocab
from .functional.evaluate import dcase2023_evaluate, dcase2024_evaluate, evaluate
from .utils.globals import (
    get_default_cache_path,
    get_default_java_path,
    get_default_tmp_path,
    set_default_cache_path,
    set_default_java_path,
    set_default_tmp_path,
)

__all__ = [
    "AACMetric",
    "BERTScoreMRefs",
    "BLEU",
    "BLEU1",
    "BLEU2",
    "BLEU3",
    "BLEU4",
    "CLAPSim",
    "CIDErD",
    "Evaluate",
    "DCASE2023Evaluate",
    "DCASE2024Evaluate",
    "FENSE",
    "FER",
    "MACE",
    "METEOR",
    "ROUGEL",
    "SBERTSim",
    "SPICE",
    "SPIDEr",
    "SPIDErFL",
    "SPIDErMax",
    "Vocab",
    "evaluate",
    "dcase2023_evaluate",
    "dcase2024_evaluate",
    "get_default_cache_path",
    "get_default_java_path",
    "get_default_tmp_path",
    "set_default_cache_path",
    "set_default_java_path",
    "set_default_tmp_path",
    "list_metrics_available",
    "load_metric",
]


def list_metrics_available() -> list[str]:
    """Returns the list of metrics that can be loaded from its name."""
    factory = _get_metric_factory_classes()
    return list(factory.keys())


def load_metric(name: str, **kwargs) -> AACMetric:
    """Load a metric class by name.

    :param name: The name of the metric.
    :param **kwargs: The optional keyword arguments passed to the metric factory.
    :returns: The Metric object built.
    """
    if not isinstance(name, str):
        raise TypeError(f"Invalid argument type {type(name)}. (expected str)")

    metrics = _instantiate_metrics_classes(name, **kwargs)
    return metrics[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import re
import subprocess
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Union

import pythonwrench as pw
from pythonwrench.semver import Version
from typing_extensions import TypeGuard

pylog = logging.getLogger(__name__)

MIN_JAVA_MAJOR_VERSION = 8
MAX_JAVA_MAJOR_VERSION = 13


def check_metric_inputs(
    candidates: Any,
    mult_references: Any,
    min_length: int = 0,
) -> None:
    """Raises ValueError if candidates and mult_references does not have a valid type and size."""

    msgs = []
    if not is_mono_sents(candidates):
        if isinstance(candidates, list) and len(candidates) > 0:
            clsname = (
                f"{candidates.__class__.__name__}[{candidates[0].__class__.__name__}]"
            )
        else:
            clsname = candidates.__class__.__name__

        msg = f"Invalid candidates type. (expected list[str], found {clsname})"
        msgs.append(msg)

    if not is_mult_sents(mult_references):
        clsname = mult_references.__class__.__name__
        msg = (
            f"Invalid mult_references type. (expected list[list[str]], found {clsname})"
        )
        msgs.append(msg)

    if len(msgs) > 0:
        raise ValueError("\n".join(msgs))

    same_len = len(candidates) == len(mult_references)
    if not same_len:
        msg = f"Invalid number of candidates ({len(candidates)}) with the number of references ({len(mult_references)})."
        raise ValueError(msg)

    at_least_1_ref_per_cand = all(len(refs) > 0 for refs in mult_references)
    if not at_least_1_ref_per_cand:
        msg = "Invalid number of references per candidate. (found at least 1 empty list of references)"
        raise ValueError(msg)

    if len(candidates) < min_length:
        msg = f"Invalid number of sentences in candidates and references. (expected at least {min_length} sentences but found {len(candidates)=})"
        raise ValueError(msg)


def check_java_path(java_path: Union[str, Path]) -> bool:
    version = _get_java_version(str(java_path))
    valid = _check_java_version(version, MIN_JAVA_MAJOR_VERSION, MAX_JAVA_MAJOR_VERSION)
    if not valid:
        msg = (
            f"Using Java version {version} is not officially supported by aac-metrics package and will not work for METEOR and SPICE metrics."
            f"(expected major version in range [{MIN_JAVA_MAJOR_VERSION}, {MAX_JAVA_MAJOR_VERSION}])"
        )
        pylog.error(msg)
    return valid


def is_mono_sents(sents: Any) -> TypeGuard[list[str]]:
    """Returns True if input is list[str] containing sentences."""
    return pw.isinstance_generic(sents, list[str])


def is_mult_sents(mult_sents: Any) -> TypeGuard[list[list[str]]]:
    """Returns True if input is list[list[str]] containing multiple sentences."""
    return pw.isinstance_generic(mult_sents, list[list[str]])


def _get_java_version(java_path: str) -> str:
    """Returns True if the java path is valid."""
    if not isinstance(java_path, str):
        msg = f"Invalid argument type {type(java_path)=}. (expected str)"
        raise TypeError(msg)

    output = "INVALID"
    try:
        output = subprocess.check_output(
            [java_path, "-version"],
            stderr=subprocess.STDOUT,
        )
        output = output.decode().strip()
        version = re.split(" |\n", output)[2][1:-1].split("_")[0]

    except (
        CalledProcessError,
        PermissionError,
        FileNotFoundError,
    ) as err:
        raise ValueError(f"Invalid java path. (from {java_path=} and found {err=})")

    except IndexError as err:
        msg = (
            f"Invalid java version. (from {java_path=} and found {output=} and {err=})"
        )
        raise ValueError(msg)

    return version


def _check_java_version(version_str: str, min_major: int, max_major: int) -> bool:
    version = Version(version_str)

    if version.major == 1 and version.minor <= 8:
        # java <= 8 use versioning "1.MAJOR.MINOR" and > 8 use "MAJOR.MINOR.MICRO"
        version_str = ".".join(map(str, (version.minor, version.micro, 0)))
        version = Version(version_str)

    return Version(f"{min_major}.0.0") <= version < Version(f"{max_major + 1}.0.0")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from typing import Callable, Literal, TypedDict, TypeVar, Union, get_args

import torch
from torch import Generator, Tensor

from aac_metrics.utils.checks import check_metric_inputs, is_mono_sents

pylog = logging.getLogger(__name__)


T = TypeVar("T")
PopStrategyName = Literal["max", "min"]
PopStrategy = Union[PopStrategyName, int]
VocabScores = TypedDict("VocabScores", {"vocab.cands": Tensor})
VocabOuts = tuple[VocabScores, VocabScores]


def vocab(
    candidates: list[str],
    mult_references: Union[list[list[str]], None],
    return_all_scores: bool = True,
    *,
    seed: Union[None, int, torch.Generator] = 1234,
    tokenizer: Callable[[str], list[str]] = str.split,
    dtype: torch.dtype = torch.float64,
    pop_strategy: PopStrategy = "max",
    verbose: int = 0,
) -> Union[VocabOuts, Tensor]:
    """Compute vocabulary statistics.

    Returns the candidate corpus vocabulary length, the references vocabulary length, the average vocabulary length for single references, and the vocabulary ratios between candidates and references.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target. Can also be None.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param seed: Random seed used to compute average vocabulary length for multiple references. defaults to 1234.
    :param tokenizer: The function used to split a sentence into tokens. defaults to str.split.
    :param dtype: Torch floating point dtype for numerical precision. defaults to torch.float64.
    :param pop_strategy: Strategy to compute average reference vocab. defaults to "max".
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    if mult_references is not None:
        check_metric_inputs(candidates, mult_references)
    elif not is_mono_sents(candidates):
        error_msg = f"Invalid candidates type. (expected list[str], found {candidates.__class__.__name__})"
        raise ValueError(error_msg)

    tok_cands = list(map(tokenizer, candidates))
    del candidates

    vocab_cands_len = _corpus_vocab_size(tok_cands, dtype)
    _, vocab_per_cand = _sent_vocab_sizes(tok_cands, dtype)
    if not return_all_scores:
        return vocab_cands_len

    sents_scores = {
        "vocab.cands": vocab_per_cand,
    }
    corpus_scores = {
        "vocab.cands": vocab_cands_len,
    }

    if mult_references is None:
        vocab_outs = corpus_scores, sents_scores
        return vocab_outs  # type: ignore

    if len(mult_references) <= 0:
        msg = f"Invalid number of references. (found {len(mult_references)} references)"
        raise ValueError(msg)
    tok_mrefs = [list(map(tokenizer, refs)) for refs in mult_references]
    del mult_references

    corpus_vocab_cands = set(token for cand in tok_cands for token in cand)
    corpus_vocab_mrefs = set(
        token for refs in tok_mrefs for ref in refs for token in ref
    )

    inter = corpus_vocab_cands.intersection(corpus_vocab_mrefs)  # True positives
    diff = corpus_vocab_mrefs.difference(corpus_vocab_cands)  # False negatives
    union = corpus_vocab_cands.union(corpus_vocab_mrefs)

    vocab_precision = len(inter) / len(corpus_vocab_cands)
    vocab_recall = len(inter) / (len(inter) + len(diff))
    vocab_f1 = 2 * vocab_precision * vocab_recall / (vocab_precision + vocab_recall)
    vocab_jaccard = len(inter) / len(union)

    vocab_mrefs_len_full = _corpus_vocab_size(
        [ref for refs in tok_mrefs for ref in refs], dtype
    )
    vocab_ratio_len_full = vocab_cands_len / vocab_mrefs_len_full

    if isinstance(seed, int):
        generator = torch.Generator().manual_seed(seed)
    else:
        generator = seed

    if pop_strategy == "max":
        num_try = max(len(refs) for refs in tok_mrefs)
    elif pop_strategy == "min":
        num_try = min(len(refs) for refs in tok_mrefs)
    elif isinstance(pop_strategy, int):
        num_try = pop_strategy
    else:
        msg = f"Invalid argument {pop_strategy=}. (expected one of {get_args(PopStrategyName)} or an integer value)"
        raise ValueError(msg)

    if verbose >= 2:
        pylog.debug(f"Found {num_try=} with {pop_strategy=}.")

    vocab_mrefs_lens = torch.empty((num_try,), dtype=dtype)

    for i in range(num_try):
        popped_refs, _ = _sample_sentences_split(tok_mrefs, generator=generator)
        vocab_mrefs_len_i = _corpus_vocab_size(popped_refs, dtype)
        vocab_mrefs_lens[i] = vocab_mrefs_len_i

    vocab_mrefs_avg = vocab_mrefs_lens.mean()
    vocab_len_ratio_avg = vocab_cands_len / vocab_mrefs_avg

    corpus_scores |= {
        "vocab.mrefs_full": vocab_mrefs_len_full,
        "vocab.ratio_full": vocab_ratio_len_full,
        "vocab.mrefs_avg": vocab_mrefs_avg,
        "vocab.ratio_avg": vocab_len_ratio_avg,
        "vocab.precision": vocab_precision,
        "vocab.recall": vocab_recall,
        "vocab.f1": vocab_f1,
        "vocab.jaccard": vocab_jaccard,
    }
    vocab_outs = corpus_scores, sents_scores
    return vocab_outs  # type: ignore


def _corpus_vocab_size(tok_sents: list[list[str]], dtype: torch.dtype) -> Tensor:
    corpus_vocab = set(token for sent in tok_sents for token in sent)
    vocab_len = torch.as_tensor(len(corpus_vocab), dtype=dtype)
    return vocab_len


def _sent_vocab_sizes(
    tok_sents: list[list[str]],
    dtype: torch.dtype,
) -> tuple[Tensor, Tensor]:
    sents_cands_vocabs = [set(sent) for sent in tok_sents]
    sent_cands_vocabs_lens = torch.as_tensor(
        list(map(len, sents_cands_vocabs)), dtype=dtype
    )
    sent_cands_vocab_len = sent_cands_vocabs_lens.mean()
    return sent_cands_vocab_len, sent_cands_vocabs_lens


def _sample_sentences_split(
    mult_sentences: list[list[T]],
    generator: Union[Generator, None] = None,
) -> tuple[list[T], list[list[T]]]:
    candidates: list[T] = []
    mult_references: list[list[T]] = []

    for sents in mult_sentences:
        idx = int(torch.randint(0, len(sents), (), generator=generator).item())
        cand = sents[idx]
        refs = [sent for i, sent in enumerate(sents) if i != idx]
        candidates.append(cand)
        mult_references.append(refs)

    return candidates, mult_references



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import importlib
import logging
import random
import sys
import unittest
from functools import partial
from pathlib import Path
from typing import Callable
from unittest import TestCase

import torch

from aac_metrics.classes.mace import MACE

pylog = logging.getLogger(__name__)


class TestMACECompatibility(TestCase):
    # Set Up methods
    @classmethod
    def setUpClass(cls) -> None:
        method = "combined"
        echecker = "echecker_clotho_audiocaps_base"
        device = "cuda" if torch.cuda.is_available() else "cpu"
        # device = "cpu"

        src_mace_fn = cls._get_src_mace()
        src_mace_fn = partial(src_mace_fn, method=method, echecker=echecker)

        cls.new_mace = MACE(
            mace_method=method,
            return_all_scores=True,
            device=device,
            verbose=2,
            echecker=echecker,
            seed=None,  # seed will be set just before testing to ensure same seed for src and new MACE
        )
        cls.src_mace = src_mace_fn

    @classmethod
    def _get_src_mace(cls) -> Callable:
        mace_dpath = Path(__file__).parent.joinpath("mace", "mace_metric")
        sys.path.append(str(mace_dpath))
        mace_module = importlib.import_module("mace")
        mace_fn = mace_module.mace
        return mace_fn

    # Tests methods
    def test_examples_mace(self) -> None:
        audio_dpath = Path(__file__).parent.joinpath("mace", "assets")
        candidates = ["a man is speaking", "rain falls"]
        mult_references = [
            [
                "a man speaks.",
                "someone speaks.",
                "a man is speaking while a bird is chirping in the background",
            ],
            ["rain is falling hard on a surface"],
        ]
        audio_paths = [
            audio_dpath.joinpath("woman_singing.wav"),
            audio_dpath.joinpath("rain.wav"),
        ]
        audio_paths = list(map(str, audio_paths))

        seed = random.randint(0, 2**16)

        random.seed(seed)
        new_outs = self.new_mace(
            candidates=candidates,
            mult_references=mult_references,
            audio_paths=audio_paths,
        )

        random.seed(seed)
        src_outs = self.src_mace(
            candidates=candidates,
            mult_references=mult_references,
            audio_paths=audio_paths,
        )

        for new_dict, src_dict in zip(new_outs, src_outs):
            assert isinstance(new_dict, dict)
            assert isinstance(src_dict, dict)
            assert set(new_dict.keys()) == set(src_dict.keys())

            equals = {k: torch.equal(new_dict[k], src_dict[k]) for k in new_dict.keys()}
            assert all(equals.values()), f"{equals=}; {new_dict=}; {src_dict=}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import TypeVar

T = TypeVar("T")


def duplicate_list(lst: list[T], sizes: list[int]) -> list[T]:
    """Duplicate elements elements of a list with the corresponding sizes.

    Example 1
    ----------
    >>> lst = ["a", "b", "c", "d", "e"]
    >>> sizes = [1, 0, 2, 1, 3]
    >>> duplicate_list(lst, sizes)
    ... ["a", "c", "c", "d", "e", "e", "e"]
    """
    if len(lst) != len(sizes):
        msg = f"Invalid arguments lengths. (found {len(lst)=} != {len(sizes)=})"
        raise ValueError(msg)

    out_size = sum(sizes)
    out: list[T] = [None for _ in range(out_size)]  # type: ignore
    curidx = 0
    for size, elt in zip(sizes, lst):
        out[curidx : curidx + size] = [elt] * size
        curidx += size
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import inspect
import logging
import time
from functools import partial
from pathlib import Path
from typing import Any, Callable, Iterable, Union

import pythonwrench as pw
import torch
from torch import Tensor, nn

from aac_metrics.functional.bert_score_mrefs import bert_score_mrefs
from aac_metrics.functional.bleu import bleu, bleu_1, bleu_2, bleu_3, bleu_4
from aac_metrics.functional.cider_d import cider_d
from aac_metrics.functional.clap_sim import clap_sim
from aac_metrics.functional.fense import fense
from aac_metrics.functional.fer import fer
from aac_metrics.functional.mace import mace
from aac_metrics.functional.meteor import meteor
from aac_metrics.functional.rouge_l import rouge_l
from aac_metrics.functional.sbert_sim import sbert_sim
from aac_metrics.functional.spice import spice
from aac_metrics.functional.spider import spider
from aac_metrics.functional.spider_fl import spider_fl
from aac_metrics.functional.spider_max import spider_max
from aac_metrics.functional.vocab import vocab
from aac_metrics.utils.checks import check_metric_inputs
from aac_metrics.utils.tokenization import preprocess_mono_sents

pylog = logging.getLogger(__name__)

METRICS_SETS: dict[str, tuple[str, ...]] = {
    # Legacy metrics for AAC
    "default": (
        "bleu_1",
        "bleu_2",
        "bleu_3",
        "bleu_4",
        "meteor",
        "rouge_l",
        "spider",  # includes cider_d, spice
    ),
    # DCASE challenge task6a metrics for 2020, 2021 and 2022
    "dcase2020": (
        "bleu_1",
        "bleu_2",
        "bleu_3",
        "bleu_4",
        "meteor",
        "rouge_l",
        "spider",  # includes cider_d, spice
    ),
    # DCASE challenge task6a metrics for 2023
    "dcase2023": (
        "meteor",
        "spider_fl",  # includes cider_d, spice, spider, fer
    ),
    # DCASE challenge task6 metrics for 2024
    "dcase2024": (
        "meteor",
        "spider_fl",  # includes cider_d, spice, spider, fer
        "fense",  # includes sbert, fer
        "vocab",
    ),
    # All metrics
    "all": (
        "bleu_1",
        "bleu_2",
        "bleu_3",
        "bleu_4",
        "meteor",
        "rouge_l",
        "fense",  # includes sbert, fer
        "spider_fl",  # includes cider_d, spice, spider, fer
        "vocab",
        "bert_score",
        "mace",
    ),
}
DEFAULT_METRICS_SET_NAME = "default"


def evaluate(
    candidates: list[str],
    mult_references: list[list[str]],
    preprocess: Union[bool, Callable[[list[str]], list[str]]] = True,
    metrics: Union[
        str, Iterable[str], Iterable[Callable[[list, list], tuple]]
    ] = DEFAULT_METRICS_SET_NAME,
    cache_path: Union[str, Path, None] = None,
    java_path: Union[str, Path, None] = None,
    tmp_path: Union[str, Path, None] = None,
    device: Union[str, torch.device, None] = "cuda_if_available",
    verbose: int = 0,
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """Evaluate candidates with multiple references with custom metrics.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param preprocess: If True, the candidates and references will be passed as input to the PTB stanford tokenizer before computing metrics. defaults to True.
    :param metrics: The name of the metric list or the explicit list of metrics to compute. defaults to "default".
    :param cache_path: The path to the external code directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_cache_path`.
    :param java_path: The path to the java executable. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_java_path`.
    :param tmp_path: Temporary directory path. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_tmp_path`.
    :param device: The PyTorch device used to run FENSE and SPIDErFL models.
        If None, it will try to detect use cuda if available. defaults to "cuda_if_available".
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple contains the corpus and sentences scores.
    """
    check_metric_inputs(candidates, mult_references)

    metrics = _instantiate_metrics_functions(
        metrics,
        cache_path=cache_path,
        java_path=java_path,
        tmp_path=tmp_path,
        device=device,
        verbose=verbose,
    )

    # Note: we use == here because preprocess is not necessary a boolean
    if preprocess == False:  # noqa: E712
        preprocess = nn.Identity()

    elif preprocess == True:  # noqa: E712
        preprocess = partial(
            preprocess_mono_sents,
            cache_path=cache_path,
            java_path=java_path,
            tmp_path=tmp_path,
            verbose=verbose,
        )

    candidates = preprocess(candidates)
    mult_references_flat, sizes = pw.flat_list_of_list(mult_references)
    mult_references_flat = preprocess(mult_references_flat)
    mult_references = pw.unflat_list_of_list(mult_references_flat, sizes)

    outs_corpus = {}
    outs_sents = {}

    for i, metric in enumerate(metrics):
        if isinstance(metric, partial):
            name = metric.func.__qualname__
        elif hasattr(metric, "__qualname__"):
            name = metric.__qualname__
        else:
            name = metric.__class__.__qualname__

        if verbose >= 1:
            pylog.info(f"[{i+1:2d}/{len(metrics):2d}] Computing {name} metric...")

        start = time.perf_counter()
        outs_corpus_i, outs_sents_i = metric(candidates, mult_references)
        end = time.perf_counter()

        if verbose >= 1:
            pylog.info(
                f"[{i+1:2d}/{len(metrics):2d}] Metric {name} computed in {end - start:.2f}s."
            )

        if __debug__:
            corpus_overlap = tuple(
                set(outs_corpus_i.keys()).intersection(outs_corpus.keys())
            )
            sents_overlap = tuple(
                set(outs_sents_i.keys()).intersection(outs_sents.keys())
            )
            if len(corpus_overlap) > 0 or len(sents_overlap) > 0:
                msg = f"Found overlapping metric outputs names. (found {corpus_overlap=} and {sents_overlap=} at least twice)"
                pw.warn_once(msg)

        outs_corpus |= outs_corpus_i
        outs_sents |= outs_sents_i

    return outs_corpus, outs_sents


def dcase2023_evaluate(
    candidates: list[str],
    mult_references: list[list[str]],
    preprocess: Union[bool, Callable[[list[str]], list[str]]] = True,
    cache_path: Union[str, Path, None] = None,
    java_path: Union[str, Path, None] = None,
    tmp_path: Union[str, Path, None] = None,
    device: Union[str, torch.device, None] = "cuda_if_available",
    verbose: int = 0,
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """Evaluate candidates with multiple references with the DCASE2023 Audio Captioning metrics.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param preprocess: If True, the candidates and references will be passed as input to the PTB stanford tokenizer before computing metrics.
        defaults to True.
    :param cache_path: The path to the external code directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_cache_path`.
    :param java_path: The path to the java executable. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_java_path`.
    :param tmp_path: Temporary directory path. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_tmp_path`.
    :param device: The PyTorch device used to run FENSE and SPIDErFL models.
        If None, it will try to detect use cuda if available. defaults to "cuda_if_available".
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple contains the corpus and sentences scores.
    """
    return evaluate(
        candidates=candidates,
        mult_references=mult_references,
        preprocess=preprocess,
        metrics="dcase2023",
        cache_path=cache_path,
        java_path=java_path,
        tmp_path=tmp_path,
        device=device,
        verbose=verbose,
    )


def dcase2024_evaluate(
    candidates: list[str],
    mult_references: list[list[str]],
    preprocess: Union[bool, Callable[[list[str]], list[str]]] = True,
    cache_path: Union[str, Path, None] = None,
    java_path: Union[str, Path, None] = None,
    tmp_path: Union[str, Path, None] = None,
    device: Union[str, torch.device, None] = "cuda_if_available",
    verbose: int = 0,
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """Evaluate candidates with multiple references with the DCASE2024 Audio Captioning metrics.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param preprocess: If True, the candidates and references will be passed as input to the PTB stanford tokenizer before computing metrics.
        defaults to True.
    :param cache_path: The path to the external code directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_cache_path`.
    :param java_path: The path to the java executable. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_java_path`.
    :param tmp_path: Temporary directory path. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_tmp_path`.
    :param device: The PyTorch device used to run FENSE and SPIDErFL models.
        If None, it will try to detect use cuda if available. defaults to "cuda_if_available".
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple contains the corpus and sentences scores.
    """
    return evaluate(
        candidates=candidates,
        mult_references=mult_references,
        preprocess=preprocess,
        metrics="dcase2024",
        cache_path=cache_path,
        java_path=java_path,
        tmp_path=tmp_path,
        device=device,
        verbose=verbose,
    )


def _instantiate_metrics_functions(
    metrics: Union[str, Iterable[str], Iterable[Callable[[list, list], tuple]]] = "all",
    *,
    cache_path: Union[str, Path, None] = None,
    java_path: Union[str, Path, None] = None,
    tmp_path: Union[str, Path, None] = None,
    device: Union[str, torch.device, None] = "cuda_if_available",
    verbose: int = 0,
) -> list[Callable]:
    if isinstance(metrics, str) and metrics in METRICS_SETS:
        metrics = METRICS_SETS[metrics]

    if isinstance(metrics, str):
        metrics = [metrics]
    else:
        metrics = list(metrics)  # type: ignore

    if not all(isinstance(metric, (str, Callable)) for metric in metrics):
        raise TypeError(
            "Invalid argument type for metrics. (expected str, Iterable[str] or Iterable[Metric])"
        )

    metric_factory = _get_metric_factory_functions(
        return_all_scores=True,
        cache_path=cache_path,
        java_path=java_path,
        tmp_path=tmp_path,
        device=device,
        verbose=verbose,
    )

    metrics_inst: list[Callable] = []
    for metric in metrics:
        if isinstance(metric, str):
            metric = metric_factory[metric]
        metrics_inst.append(metric)
    return metrics_inst


def _get_metric_factory_functions(
    **kwargs,
) -> dict[str, Callable[[list[str], list[list[str]]], Any]]:
    functions = {
        "bert_score": bert_score_mrefs,
        "bleu": bleu,
        "bleu_1": bleu_1,
        "bleu_2": bleu_2,
        "bleu_3": bleu_3,
        "bleu_4": bleu_4,
        "clap_sim": clap_sim,
        "cider_d": cider_d,
        "fer": fer,
        "fense": fense,
        "mace": mace,
        "meteor": meteor,
        "rouge_l": rouge_l,
        "sbert_sim": sbert_sim,
        "spice": spice,
        "spider": spider,
        "spider_max": spider_max,
        "spider_fl": spider_fl,
        "vocab": vocab,
    }
    factory = {}
    for name, fn in functions.items():
        argnames = get_argnames(fn)
        fn_kwargs = {k: v for k, v in kwargs.items() if k in argnames}
        factory[name] = partial(fn, **fn_kwargs)
    return factory


def get_argnames(fn: Callable) -> list[str]:
    """Get arguments names of a method, function or callable object."""
    if inspect.ismethod(fn):
        # If method, remove 'self' arg
        argnames = fn.__code__.co_varnames[1:]  # type: ignore
    elif inspect.isfunction(fn):
        argnames = fn.__code__.co_varnames
    else:
        argnames = fn.__call__.__code__.co_varnames  # type: ignore

    argnames = list(argnames)
    return argnames



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional, Union, get_args

import torch
from torch import Tensor
from transformers.models.auto.tokenization_auto import AutoTokenizer

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.clap_sim import CLAP, DEFAULT_CLAP_SIM_MODEL, _load_clap
from aac_metrics.functional.fer import (
    DEFAULT_FER_MODEL,
    BERTFlatClassifier,
    _load_echecker_and_tokenizer,
)
from aac_metrics.functional.mace import MACEMethod, MACEOuts, mace
from aac_metrics.utils.globals import _get_device


class MACE(AACMetric[Union[MACEOuts, Tensor]]):
    """Multimodal Audio-Caption Evaluation class (MACE).

    MACE is a metric designed for evaluating automated audio captioning (AAC) systems.
    Unlike metrics that compare machine-generated captions solely to human references, MACE uses both audio and text to improve evaluation.
    By integrating both audio and text, it produces assessments that align better with human judgments.

    The implementation is based on the mace original implementation (original author have accepted to include their code in aac-metrics under the MIT license).
    Note: Instances of this class are not pickable.

    - Paper: https://arxiv.org/pdf/2411.00321
    - Original author: Satvik Dixit
    - Original implementation: https://github.com/satvik-dixit/mace/tree/main

    For more information, see :func:`~aac_metrics.functional.mace.mace`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = -1.0
    max_value = 1.0

    def __init__(
        self,
        return_all_scores: bool = True,
        *,
        mace_method: MACEMethod = "text",
        penalty: float = 0.3,
        # CLAP args
        clap_model: Union[str, CLAP] = DEFAULT_CLAP_SIM_MODEL,
        seed: Optional[int] = 42,
        # FER args
        echecker: Union[str, BERTFlatClassifier] = DEFAULT_FER_MODEL,
        echecker_tokenizer: Optional[AutoTokenizer] = None,
        error_threshold: float = 0.97,
        device: Union[str, torch.device, None] = "cuda_if_available",
        batch_size: Optional[int] = 32,
        reset_state: bool = True,
        return_probs: bool = False,
        # Other args
        verbose: int = 0,
    ) -> None:
        if mace_method not in get_args(MACEMethod):
            msg = f"Invalid argument {mace_method=}. (expected one of {get_args(MACEMethod)})"
            raise ValueError(msg)

        device = _get_device(device)
        clap_model = _load_clap(
            clap_model=clap_model,
            device=device,
            reset_state=reset_state,
        )
        echecker, echecker_tokenizer = _load_echecker_and_tokenizer(
            echecker=echecker,
            echecker_tokenizer=echecker_tokenizer,
            device=device,
            reset_state=reset_state,
            verbose=verbose,
        )

        super().__init__()
        self._return_all_scores = return_all_scores
        self._mace_method: MACEMethod = mace_method
        self._penalty = penalty
        self._clap_model = clap_model
        self._seed = seed
        self._echecker = echecker
        self._echecker_tokenizer = echecker_tokenizer
        self._error_threshold = error_threshold
        self._device = device
        self._batch_size = batch_size
        self._reset_state = reset_state
        self._return_probs = return_probs
        self._verbose = verbose

        self._candidates = []
        self._mult_references = []
        self._audio_paths = []

    def compute(self) -> Union[MACEOuts, Tensor]:
        return mace(
            candidates=self._candidates,
            mult_references=self._mult_references,
            audio_paths=self._audio_paths,
            return_all_scores=self._return_all_scores,
            mace_method=self._mace_method,
            clap_model=self._clap_model,
            seed=self._seed,
            echecker=self._echecker,
            echecker_tokenizer=self._echecker_tokenizer,
            error_threshold=self._error_threshold,
            device=self._device,
            batch_size=self._batch_size,
            reset_state=self._reset_state,
            return_probs=self._return_probs,
            penalty=self._penalty,
            verbose=self._verbose,
        )

    def extra_repr(self) -> str:
        if isinstance(self._clap_model, str):
            model_name = self._clap_model
        else:
            model_name = self._clap_model.__class__.__name__

        hparams = {"clap_model": model_name}
        repr_ = ", ".join(f"{k}={v}" for k, v in hparams.items())
        return repr_

    def get_output_names(self) -> tuple[str, ...]:
        return ("mace", "fer", "clap_sim")

    def reset(self) -> None:
        self._candidates = []
        self._mult_references = []
        self._audio_paths = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references: Optional[list[list[str]]] = None,
        audio_paths: Optional[list[str]] = None,
    ) -> None:
        self._candidates += candidates

        if self._mace_method == "audio":
            if mult_references is None:
                msg = f"Invalid argument {mult_references=} with {self._mace_method=}."
                raise ValueError(msg)

            self._mult_references += mult_references

        elif self._mace_method == "text":
            if audio_paths is None:
                msg = f"Invalid argument {audio_paths=} with {self._mace_method=}."
                raise ValueError(msg)

            self._audio_paths += audio_paths

        elif self._mace_method == "combined":
            if mult_references is None:
                msg = f"Invalid argument {mult_references=} with {self._mace_method=}."
                raise ValueError(msg)
            if audio_paths is None:
                msg = f"Invalid argument {audio_paths=} with {self._mace_method=}."
                raise ValueError(msg)

            self._mult_references += mult_references
            self._audio_paths += audio_paths

        else:
            msg = f"Invalid value {self._mace_method=}. (expected one of {get_args(MACEMethod)})"
            raise ValueError(msg)

    def __getstate__(self) -> bytes:  # type: ignore
        raise RuntimeError(f"{self.__class__.__name__} is not pickable.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import os.path as osp
import subprocess
import tempfile
import time
from pathlib import Path
from typing import Any, Hashable, Iterable, Optional, Union

import pythonwrench as pw

from aac_metrics.utils.checks import check_java_path, is_mono_sents
from aac_metrics.utils.globals import _get_cache_path, _get_java_path, _get_tmp_path

pylog = logging.getLogger(__name__)


# Path to the stanford corenlp jar
FNAME_STANFORD_CORENLP_3_4_1_JAR = osp.join(
    "aac-metrics",
    "stanford_nlp",
    "stanford-corenlp-3.4.1.jar",
)
# Punctuations to be removed from the sentences
PTB_PUNCTUATIONS = (
    "''",
    "'",
    "``",
    "`",
    "-LRB-",
    "-RRB-",
    "-LCB-",
    "-RCB-",
    ".",
    "?",
    "!",
    ",",
    ":",
    "-",
    "--",
    "...",
    ";",
)


def ptb_tokenize_batch(
    sentences: Iterable[str],
    audio_ids: Optional[Iterable[Hashable]] = None,
    cache_path: Union[str, Path, None] = None,
    java_path: Union[str, Path, None] = None,
    tmp_path: Union[str, Path, None] = None,
    punctuations: Iterable[str] = PTB_PUNCTUATIONS,
    normalize_apostrophe: bool = False,
    verbose: int = 0,
) -> list[list[str]]:
    """Use PTB Tokenizer to process sentences. Should be used only with all the sentences of a subset due to slow computation.

    :param sentences: The sentences to tokenize.
    :param audio_ids: The optional audio names for the PTB Tokenizer program. None will use the audio index as name. defaults to None.
    :param cache_path: The path to the external directory containing the JAR program. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_cache_path`.
    :param java_path: The path to the java executable. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_java_path`.
    :param tmp_path: The path to a temporary directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_tmp_path`.
    :param normalize_apostrophe: If True, add apostrophes for French language. defaults to False.
    :param verbose: The verbose level. defaults to 0.
    :returns: The sentences tokenized as list[list[str]].
    """
    # Originally based on https://github.com/audio-captioning/caption-evaluation-tools/blob/c1798df4c91e29fe689b1ccd4ce45439ec966417/caption/pycocoevalcap/tokenizer/ptbtokenizer.py#L30

    sentences = list(sentences)

    if not is_mono_sents(sentences):
        raise ValueError("Invalid argument sentences. (not a list[str] of sentences)")

    if len(sentences) == 0:
        return []

    cache_path = _get_cache_path(cache_path)
    java_path = _get_java_path(java_path)
    tmp_path = _get_tmp_path(tmp_path)
    punctuations = list(punctuations)

    stanford_fpath = osp.join(cache_path, FNAME_STANFORD_CORENLP_3_4_1_JAR)

    # Sanity checks
    if __debug__:
        newlines_count = sum(sent.count("\n") for sent in sentences)
        if newlines_count > 0:
            raise ValueError(
                f"Invalid argument sentences for tokenization. (found {newlines_count} newlines character '\\n')"
            )

        if not osp.isdir(cache_path):
            raise RuntimeError(f"Cannot find cache directory at {cache_path=}.")
        if not osp.isdir(tmp_path):
            raise RuntimeError(f"Cannot find tmp directory at {tmp_path=}.")
        if not osp.isfile(stanford_fpath):
            raise FileNotFoundError(
                f"Cannot find JAR file '{stanford_fpath}' for tokenization. Maybe run 'aac-metrics-download' or specify another 'cache_path' directory."
            )
        if not check_java_path(java_path):
            raise RuntimeError(
                f"Invalid Java executable to tokenize sentences. ({java_path})"
            )

    start_time = time.perf_counter()
    if verbose >= 2:
        pylog.debug(
            f"Start executing {FNAME_STANFORD_CORENLP_3_4_1_JAR} JAR file for tokenization. ({len(sentences)=})"
        )

    cmd = [
        java_path,
        "-cp",
        stanford_fpath,
        "edu.stanford.nlp.process.PTBTokenizer",
        "-preserveLines",
        "-lowerCase",
    ]

    if audio_ids is None:
        audio_ids = list(range(len(sentences)))
    else:
        audio_ids = list(audio_ids)

    if len(audio_ids) != len(sentences):
        raise ValueError(
            f"Invalid number of audio ids ({len(audio_ids)}) with sentences len={len(sentences)}."
        )

    sentences = "\n".join(sentences)
    if normalize_apostrophe:
        replaces = {
            " s ": " s'",
            "'": "' ",
            "'  ": "' ",
            " '": "'",
        }
        for old, new in replaces.items():
            sentences = sentences.replace(old, new)

    tmp_file = tempfile.NamedTemporaryFile(
        delete=False,
        dir=tmp_path,
        prefix="ptb_sentences_",
        suffix=".txt",
    )
    tmp_file.write(sentences.encode())
    tmp_file.close()

    cmd.append(osp.basename(tmp_file.name))
    p_tokenizer = subprocess.Popen(
        cmd,
        cwd=tmp_path,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL if verbose <= 2 else None,
    )
    encoded_sentences = sentences.rstrip().encode()
    token_lines = p_tokenizer.communicate(input=encoded_sentences)[0]
    token_lines = token_lines.decode()
    lines = token_lines.split("\n")
    # remove temp file
    os.remove(tmp_file.name)

    if len(audio_ids) != len(lines):
        msg = (
            f"PTB tokenize error: expected {len(audio_ids)} lines in output file but found {len(lines)}."
            f"Maybe check if there is any newline character '\\n' in your sentences or disable preprocessing tokenization."
        )
        raise RuntimeError(msg)

    outs: Any = [None for _ in range(len(lines))]
    for k, line in zip(audio_ids, lines):
        tokenized_caption = [
            w for w in line.rstrip().split(" ") if w not in punctuations
        ]
        outs[k] = tokenized_caption

    msg = "INTERNAL ERROR: PTB tokenizer output is invalid."
    assert all(out is not None for out in outs), msg

    if verbose >= 2:
        duration = time.perf_counter() - start_time
        pylog.debug(f"Tokenization finished in {duration:.2f}s.")

    return outs


def preprocess_mono_sents(
    sentences: list[str],
    cache_path: Union[str, Path, None] = None,
    java_path: Union[str, Path, None] = None,
    tmp_path: Union[str, Path, None] = None,
    punctuations: Iterable[str] = PTB_PUNCTUATIONS,
    normalize_apostrophe: bool = False,
    verbose: int = 0,
) -> list[str]:
    """Tokenize sentences using PTB Tokenizer then merge them by space.

    .. warning::
        PTB tokenizer is a java program that takes a list[str] as input, so calling several times this function is slow on list[list[str]].

        If you want to process multiple sentences (list[list[str]]), use :func:`~aac_metrics.utils.tokenization.preprocess_mult_sents` instead.

    :param sentences: The list of sentences to process.
    :param cache_path: The path to the external code directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_cache_path`.
    :param java_path: The path to the java executable. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_java_path`.
    :param tmp_path: Temporary directory path. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_tmp_path`.
    :param punctuations: Set of punctuations to remove. defaults to PTB_PUNCTUATIONS.
    :param normalize_apostrophe: If True, add apostrophes for French language. defaults to False.
    :param verbose: The verbose level. defaults to 0.
    :returns: The sentences processed by the tokenizer.
    """
    tok_sents = ptb_tokenize_batch(
        sentences=sentences,
        audio_ids=None,
        cache_path=cache_path,
        java_path=java_path,
        tmp_path=tmp_path,
        punctuations=punctuations,
        normalize_apostrophe=normalize_apostrophe,
        verbose=verbose,
    )
    sentences = [" ".join(sent) for sent in tok_sents]
    return sentences


def preprocess_mult_sents(
    mult_sentences: list[list[str]],
    cache_path: Union[str, Path, None] = None,
    java_path: Union[str, Path, None] = None,
    tmp_path: Union[str, Path, None] = None,
    punctuations: Iterable[str] = PTB_PUNCTUATIONS,
    normalize_apostrophe: bool = False,
    verbose: int = 0,
) -> list[list[str]]:
    """Tokenize multiple sentences using PTB Tokenizer with only one call then merge them by space.

    :param mult_sentences: The list of list of sentences to process.
    :param cache_path: The path to the external code directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_cache_path`.
    :param java_path: The path to the java executable. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_java_path`.
    :param tmp_path: Temporary directory path. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_tmp_path`.
    :param normalize_apostrophe: If True, add apostrophes for French language. defaults to False.
    :param verbose: The verbose level. defaults to 0.
    :returns: The multiple sentences processed by the tokenizer.
    """
    flatten_sents, sizes = pw.flat_list_of_list(mult_sentences)
    flatten_sents = preprocess_mono_sents(
        sentences=flatten_sents,
        cache_path=cache_path,
        java_path=java_path,
        tmp_path=tmp_path,
        punctuations=punctuations,
        normalize_apostrophe=normalize_apostrophe,
        verbose=verbose,
    )
    mult_sentences = pw.unflat_list_of_list(flatten_sents, sizes)
    return mult_sentences



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import random
from typing import Literal, Optional, TypedDict, Union, get_args

import numpy as np
import torch
from msclap import CLAP
from torch import Tensor
from torch.nn import CosineSimilarity

from aac_metrics.utils.globals import _get_device

pylog = logging.getLogger(__name__)

DEFAULT_CLAP_SIM_MODEL = "MS-CLAP-2023"

CLAPMethod = Literal["audio", "text"]

CLAPScores = TypedDict("CLAPScores", {"clap_sim": Tensor})
CLAPOuts = tuple[CLAPScores, CLAPScores]


def clap_sim(
    candidates: list[str],
    mult_references: Optional[list[list[str]]] = None,
    audio_paths: Optional[list[str]] = None,
    return_all_scores: bool = True,
    *,
    clap_method: CLAPMethod = "text",
    clap_model: Union[str, CLAP] = DEFAULT_CLAP_SIM_MODEL,
    device: Union[str, torch.device, None] = "cuda_if_available",
    batch_size: Optional[int] = 32,
    reset_state: bool = True,
    seed: Optional[int] = 42,
    verbose: int = 0,
) -> Union[Tensor, CLAPOuts]:
    """Cosine-similarity of the Contrastive Language-Audio Pretraining (CLAP) embeddings.

    The implementation is based on the msclap pypi package.

    - Paper: https://arxiv.org/pdf/2411.00321
    - msclap package: https://pypi.org/project/msclap/

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target when method is "text". defaults to None.
    :param audio_paths: Audio filepaths required when method is "audio". defaults to None.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param clap_method: The method used to encode the sentences. Can be "text" or "audio". defaults to "text".
    :param clap_model: The CLAP model used to extract sentence embeddings for cosine-similarity. defaults to "2023".
    :param device: The PyTorch device used to run MACE models. If "cuda_if_available", it will use cuda if available. defaults to "cuda_if_available".
    :param batch_size: The batch size of the CLAP models. defaults to 32.
    :param reset_state: If True, reset the state of the PyTorch global generator after the initialization of the pre-trained models. defaults to True.
    :param seed: Optional seed to make CLAP-sim scores deterministic when using clap_method="audio" on large audio files. defaults to 42.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    # Init models
    clap_model = _load_clap(clap_model, device, reset_state)
    cands_embs = _encode_sents_clap(clap_model, candidates, batch_size)

    if clap_method == "text":
        if mult_references is None:
            msg = f"Invalid arguments {clap_method=} with {mult_references=}."
            raise ValueError(msg)

        rng_ids = [0]
        for refs in mult_references:
            rng_ids.append(rng_ids[-1] + len(refs))
        flat_references = [ref for refs in mult_references for ref in refs]
        mrefs_embs = _encode_sents_clap(clap_model, flat_references, batch_size)

    elif clap_method == "audio":
        if audio_paths is None:
            msg = f"Invalid arguments {clap_method=} with {audio_paths=}."
            raise ValueError(msg)

        rng_ids = [i for i in range(len(audio_paths) + 1)]
        if seed is not None:
            random.seed(seed)
        mrefs_embs = _encode_audios_clap(clap_model, audio_paths, batch_size)

    else:
        msg = (
            f"Invalid argument {clap_method=}. (expected one of {get_args(CLAPMethod)})"
        )
        raise ValueError(msg)

    clap_sim_scores = [
        _cosine_similarity(cands_embs[i], mrefs_embs[rng_ids[i] : rng_ids[i + 1]])
        .mean()
        .item()
        for i in range(len(cands_embs))
    ]
    clap_sim_scores = np.array(clap_sim_scores)

    # Aggregate and return
    clap_sim_score = clap_sim_scores.mean()
    clap_sim_score = torch.as_tensor(clap_sim_score)
    clap_sim_scores = torch.as_tensor(clap_sim_scores)

    if return_all_scores:
        clap_sim_outs_corpus = {
            "clap_sim": clap_sim_score,
        }
        clap_sim_outs_sents = {
            "clap_sim": clap_sim_scores,
        }
        clap_outs = clap_sim_outs_corpus, clap_sim_outs_sents
        return clap_outs  # type: ignore
    else:
        return clap_sim_score


def _cosine_similarity(input: Tensor, target: Tensor) -> Tensor:
    cos = CosineSimilarity(dim=-1, eps=1e-6)
    return cos(input.unsqueeze(0), target)


def _load_clap(
    clap_model: Union[str, CLAP] = DEFAULT_CLAP_SIM_MODEL,
    device: Union[str, torch.device, None] = "cuda_if_available",
    reset_state: bool = True,
) -> CLAP:
    state = torch.random.get_rng_state()

    device = _get_device(device)
    if isinstance(clap_model, str):
        use_cuda = device is not None and device.type == "cuda"
        clap_model = CLAP(version="2023", use_cuda=use_cuda)

    if reset_state:
        torch.random.set_rng_state(state)
    return clap_model


@torch.no_grad()
def _encode_sents_clap(
    clap_model: CLAP,
    sents: list[str],
    batch_size: Optional[int] = 32,
) -> Tensor:
    if batch_size is None:
        batch_size = len(sents)

    clap_embeddings = []
    for i in range(0, len(sents), batch_size):
        sents_batch = sents[i : i + batch_size]
        clap_embeddings_batch = clap_model.get_text_embeddings(sents_batch)
        clap_embeddings.append(clap_embeddings_batch)
    clap_embeddings = torch.vstack(clap_embeddings)
    return clap_embeddings


@torch.no_grad()
def _encode_audios_clap(
    clap_model: CLAP,
    audio_paths: list[str],
    batch_size: Optional[int] = 32,
) -> Tensor:
    audio_paths = list(map(str, audio_paths))
    if batch_size is None:
        batch_size = len(audio_paths)

    clap_embeddings = []
    for i in range(0, len(audio_paths), batch_size):
        audio_paths_batch = audio_paths[i : i + batch_size]
        clap_embeddings_batch = clap_model.get_audio_embeddings(audio_paths_batch)
        clap_embeddings.append(clap_embeddings_batch)
    clap_embeddings = torch.vstack(clap_embeddings)
    return clap_embeddings
